@@ -2,8 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2008-2009  Nokia Corporation
- *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2008-2010  Nokia Corporation
+ *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -28,6 +28,8 @@
 
 #include <stdlib.h>
 #include <stdio.h>
+#include <unistd.h>
+#include <fcntl.h>
 #include <stdint.h>
 #include <string.h>
 #include <glib.h>
@@ -37,6 +39,11 @@
 #include "logging.h"
 #include "telephony.h"
 
+/* SSC D-Bus definitions */
+#define SSC_DBUS_NAME  "com.nokia.phone.SSC"
+#define SSC_DBUS_IFACE "com.nokia.phone.SSC"
+#define SSC_DBUS_PATH  "/com/nokia/phone/SSC"
+
 /* libcsnet D-Bus definitions */
 #define NETWORK_BUS_NAME		"com.nokia.phone.net"
 #define NETWORK_INTERFACE		"Phone.Net"
@@ -83,12 +90,23 @@ enum network_alpha_tag_name_type {
 	NETWORK_NITZ_FULL_OPER_NAME,
 };
 
+#define TELEPHONY_MAEMO_PATH		"/com/nokia/MaemoTelephony"
+#define TELEPHONY_MAEMO_INTERFACE	"com.nokia.MaemoTelephony"
+
+#define CALLERID_BASE		"/var/lib/bluetooth/maemo-callerid-"
+#define ALLOWED_FLAG_FILE	"/var/lib/bluetooth/maemo-callerid-allowed"
+#define RESTRICTED_FLAG_FILE	"/var/lib/bluetooth/maemo-callerid-restricted"
+#define NONE_FLAG_FILE		"/var/lib/bluetooth/maemo-callerid-none"
+
+static uint32_t callerid = 0;
+
 /* CSD CALL plugin D-Bus definitions */
 #define CSD_CALL_BUS_NAME	"com.nokia.csd.Call"
 #define CSD_CALL_INTERFACE	"com.nokia.csd.Call"
 #define CSD_CALL_INSTANCE	"com.nokia.csd.Call.Instance"
 #define CSD_CALL_CONFERENCE	"com.nokia.csd.Call.Conference"
 #define CSD_CALL_PATH		"/com/nokia/csd/call"
+#define CSD_CALL_CONFERENCE_PATH "/com/nokia/csd/call/conference"
 
 /* Call status values as exported by the CSD CALL plugin */
 #define CSD_CALL_STATUS_IDLE			0
@@ -173,6 +191,9 @@ static DBusConnection *connection = NULL;
 
 static GSList *calls = NULL;
 
+/* Reference count for determining the call indicator status */
+static GSList *active_calls = NULL;
+
 static char *msisdn = NULL;	/* Subscriber number */
 static char *vmbx = NULL;	/* Voice mailbox number */
 
@@ -181,6 +202,8 @@ static int battchg_cur = -1;	/* "battery.charge_level.current" */
 static int battchg_last = -1;	/* "battery.charge_level.last_full" */
 static int battchg_design = -1;	/* "battery.charge_level.design" */
 
+static gboolean get_calls_active = FALSE;
+
 static gboolean events_enabled = FALSE;
 
 /* Supported set of call hold operations */
@@ -196,15 +219,18 @@ static int response_and_hold = -1;
 
 static char *last_dialed_number = NULL;
 
+/* Timer for tracking call creation requests */
+static guint create_request_timer = 0;
+
 static struct indicator maemo_indicators[] =
 {
-	{ "battchg",	"0-5",	5 },
-	{ "signal",	"0-5",	5 },
-	{ "service",	"0,1",	1 },
-	{ "call",	"0,1",	0 },
-	{ "callsetup",	"0-3",	0 },
-	{ "callheld",	"0-2",	0 },
-	{ "roam",	"0,1",	0 },
+	{ "battchg",	"0-5",	5,	TRUE },
+	{ "signal",	"0-5",	0,	TRUE },
+	{ "service",	"0,1",	0,	TRUE },
+	{ "call",	"0,1",	0,	TRUE },
+	{ "callsetup",	"0-3",	0,	TRUE },
+	{ "callheld",	"0-2",	0,	FALSE },
+	{ "roam",	"0,1",	0,	TRUE },
 	{ NULL }
 };
 
@@ -243,7 +269,24 @@ static struct csd_call *find_call(const char *path)
 	return NULL;
 }
 
-static struct csd_call *find_active_call(void)
+static struct csd_call *find_non_held_call(void)
+{
+	GSList *l;
+
+	for (l = calls; l != NULL; l = l->next) {
+		struct csd_call *call = l->data;
+
+		if (call->status == CSD_CALL_STATUS_IDLE)
+			continue;
+
+		if (call->status != CSD_CALL_STATUS_HOLD)
+			return call;
+	}
+
+	return NULL;
+}
+
+static struct csd_call *find_non_idle_call(void)
 {
 	GSList *l;
 
@@ -271,6 +314,26 @@ static struct csd_call *find_call_with_status(int status)
 	return NULL;
 }
 
+static int release_conference(void)
+{
+	DBusMessage *msg;
+
+	debug("telephony-maemo: releasing conference call");
+
+	msg = dbus_message_new_method_call(CSD_CALL_BUS_NAME,
+						CSD_CALL_CONFERENCE_PATH,
+						CSD_CALL_INSTANCE,
+						"Release");
+	if (!msg) {
+		error("Unable to allocate new D-Bus message");
+		return -ENOMEM;
+	}
+
+	g_dbus_send_message(connection, msg);
+
+	return 0;
+}
+
 static int release_call(struct csd_call *call)
 {
 	DBusMessage *msg;
@@ -410,6 +473,17 @@ static int call_transfer(void)
 	return 0;
 }
 
+static int number_type(const char *number)
+{
+	if (number == NULL)
+		return NUMBER_TYPE_TELEPHONY;
+
+	if (number[0] == '+' || strncmp(number, "00", 2) == 0)
+		return NUMBER_TYPE_INTERNATIONAL;
+
+	return NUMBER_TYPE_TELEPHONY;
+}
+
 void telephony_device_connected(void *telephony_device)
 {
 	struct csd_call *coming;
@@ -420,10 +494,10 @@ void telephony_device_connected(void *telephony_device)
 	if (coming) {
 		if (find_call_with_status(CSD_CALL_STATUS_ACTIVE))
 			telephony_call_waiting_ind(coming->number,
-							NUMBER_TYPE_TELEPHONY);
+						number_type(coming->number));
 		else
 			telephony_incoming_call_ind(coming->number,
-							NUMBER_TYPE_TELEPHONY);
+						number_type(coming->number));
 	}
 }
 
@@ -464,8 +538,12 @@ void telephony_last_dialed_number_req(void *telephony_device)
 void telephony_terminate_call_req(void *telephony_device)
 {
 	struct csd_call *call;
+	int err;
+
+	call = find_call_with_status(CSD_CALL_STATUS_ACTIVE);
+	if (!call)
+		call = find_non_idle_call();
 
-	call = find_active_call();
 	if (!call) {
 		error("No active call");
 		telephony_terminate_call_rsp(telephony_device,
@@ -473,7 +551,12 @@ void telephony_terminate_call_req(void *telephony_device)
 		return;
 	}
 
-	if (release_call(call) < 0)
+	if (call->conference)
+		err = release_conference();
+	else
+		err = release_call(call);
+
+	if (err < 0)
 		telephony_terminate_call_rsp(telephony_device,
 						CME_ERROR_AG_FAILURE);
 	else
@@ -560,7 +643,7 @@ static const char *memory_dial_lookup(int location)
 
 void telephony_dial_number_req(void *telephony_device, const char *number)
 {
-	uint32_t flags = CALL_FLAG_NONE;
+	uint32_t flags = callerid;
 	int ret;
 
 	debug("telephony-maemo: dial request to %s", number);
@@ -624,7 +707,7 @@ void telephony_subscriber_number_req(void *telephony_device)
 	debug("telephony-maemo: subscriber number request");
 	if (msisdn)
 		telephony_subscriber_number_ind(msisdn,
-						NUMBER_TYPE_TELEPHONY,
+						number_type(msisdn),
 						SUBSCRIBER_SERVICE_VOICE);
 	telephony_subscriber_number_rsp(telephony_device, CME_ERROR_NONE);
 }
@@ -691,7 +774,7 @@ void telephony_list_current_calls_req(void *telephony_device)
 		telephony_list_current_call_ind(i, direction, status,
 						CALL_MODE_VOICE, multiparty,
 						call->number,
-						NUMBER_TYPE_TELEPHONY);
+						number_type(call->number));
 	}
 
 	telephony_list_current_calls_rsp(telephony_device, CME_ERROR_NONE);
@@ -831,6 +914,14 @@ void telephony_key_press_req(void *telephony_device, const char *keys)
 		telephony_key_press_rsp(telephony_device, CME_ERROR_NONE);
 }
 
+void telephony_voice_dial_req(void *telephony_device, gboolean enable)
+{
+	debug("telephony-maemo: got %s voice dial request",
+			enable ? "enable" : "disable");
+
+	telephony_voice_dial_rsp(telephony_device, CME_ERROR_NOT_SUPPORTED);
+}
+
 static void handle_incoming_call(DBusMessage *msg)
 {
 	const char *number, *call_path;
@@ -861,10 +952,10 @@ static void handle_incoming_call(DBusMessage *msg)
 
 	if (find_call_with_status(CSD_CALL_STATUS_ACTIVE))
 		telephony_call_waiting_ind(call->number,
-						NUMBER_TYPE_TELEPHONY);
+						number_type(call->number));
 	else
 		telephony_incoming_call_ind(call->number,
-						NUMBER_TYPE_TELEPHONY);
+						number_type(call->number));
 }
 
 static void handle_outgoing_call(DBusMessage *msg)
@@ -895,14 +986,38 @@ static void handle_outgoing_call(DBusMessage *msg)
 	g_free(last_dialed_number);
 	last_dialed_number = g_strdup(number);
 
+	if (create_request_timer) {
+		g_source_remove(create_request_timer);
+		create_request_timer = 0;
+	}
+}
+
+static gboolean create_timeout(gpointer user_data)
+{
+	telephony_update_indicator(maemo_indicators, "callsetup",
+					EV_CALLSETUP_INACTIVE);
+	create_request_timer = 0;
+	return FALSE;
+}
+
+static void handle_create_requested(DBusMessage *msg)
+{
+	debug("Call.CreateRequested()");
+
+	if (create_request_timer)
+		g_source_remove(create_request_timer);
+
+	create_request_timer = g_timeout_add_seconds(5, create_timeout, NULL);
+
 	telephony_update_indicator(maemo_indicators, "callsetup",
 					EV_CALLSETUP_OUTGOING);
 }
 
 static void handle_call_status(DBusMessage *msg, const char *call_path)
 {
-	struct csd_call *call, *active_call;
+	struct csd_call *call;
 	dbus_uint32_t status, cause_type, cause;
+	int callheld = telephony_get_indicator(maemo_indicators, "callheld");
 
 	if (!dbus_message_get_args(msg, NULL,
 					DBUS_TYPE_UINT32, &status,
@@ -928,7 +1043,12 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 	debug("Call %s changed from %s to %s", call_path,
 		call_status_str[call->status], call_status_str[status]);
 
-	active_call = find_call_with_status(CSD_CALL_STATUS_ACTIVE);
+	if (call->status == (int) status) {
+		debug("Ignoring CSD Call state change to existing state");
+		return;
+	}
+
+	call->status = (int) status;
 
 	switch (status) {
 	case CSD_CALL_STATUS_IDLE:
@@ -938,9 +1058,7 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 							EV_CALLSETUP_INACTIVE);
 			if (!call->originating)
 				telephony_calling_stopped_ind();
-		} else
-			telephony_update_indicator(maemo_indicators, "call",
-							EV_CALL_INACTIVE);
+		}
 
 		g_free(call->number);
 		call->number = NULL;
@@ -973,10 +1091,6 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 	case CSD_CALL_STATUS_ACTIVE:
 		if (call->on_hold) {
 			call->on_hold = FALSE;
-			/* Set the status already here since otherwise
-			 * find_call_with_status will match this call even
-			 * though it shouldn't */
-			call->status = (int) status;
 			if (find_call_with_status(CSD_CALL_STATUS_HOLD))
 				telephony_update_indicator(maemo_indicators,
 							"callheld",
@@ -986,8 +1100,17 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 							"callheld",
 							EV_CALLHELD_NONE);
 		} else {
-			telephony_update_indicator(maemo_indicators, "call",
-							EV_CALL_ACTIVE);
+			if (!g_slist_find(active_calls, call))
+				active_calls = g_slist_prepend(active_calls, call);
+			if (g_slist_length(active_calls) == 1)
+				telephony_update_indicator(maemo_indicators,
+								"call",
+								EV_CALL_ACTIVE);
+			/* Upgrade callheld status if necessary */
+			if (callheld == EV_CALLHELD_ON_HOLD)
+				telephony_update_indicator(maemo_indicators,
+							"callheld",
+							EV_CALLHELD_MULTIPLE);
 			telephony_update_indicator(maemo_indicators,
 							"callsetup",
 							EV_CALLSETUP_INACTIVE);
@@ -997,14 +1120,17 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 		}
 		break;
 	case CSD_CALL_STATUS_MO_RELEASE:
-		break;
 	case CSD_CALL_STATUS_MT_RELEASE:
+		active_calls = g_slist_remove(active_calls, call);
+		if (g_slist_length(active_calls) == 0)
+			telephony_update_indicator(maemo_indicators, "call",
+							EV_CALL_INACTIVE);
 		break;
 	case CSD_CALL_STATUS_HOLD_INITIATED:
 		break;
 	case CSD_CALL_STATUS_HOLD:
 		call->on_hold = TRUE;
-		if (active_call && active_call != call)
+		if (find_non_held_call())
 			telephony_update_indicator(maemo_indicators,
 							"callheld",
 							EV_CALLHELD_MULTIPLE);
@@ -1018,16 +1144,16 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 	case CSD_CALL_STATUS_RECONNECT_PENDING:
 		break;
 	case CSD_CALL_STATUS_TERMINATED:
-		if (call->on_hold) {
-			/* Set the status already here since otherwise
-			 * find_call_with_status will match this call even
-			 * though it shouldn't */
-			call->status = (int) status;
-			if (!find_call_with_status(CSD_CALL_STATUS_HOLD))
-				telephony_update_indicator(maemo_indicators,
-								"callheld",
-								EV_CALLHELD_NONE);
-		}
+		if (call->on_hold &&
+				!find_call_with_status(CSD_CALL_STATUS_HOLD))
+			telephony_update_indicator(maemo_indicators,
+							"callheld",
+							EV_CALLHELD_NONE);
+		else if (callheld == EV_CALLHELD_MULTIPLE &&
+				find_call_with_status(CSD_CALL_STATUS_HOLD))
+			telephony_update_indicator(maemo_indicators,
+							"callheld",
+							EV_CALLHELD_ON_HOLD);
 		break;
 	case CSD_CALL_STATUS_SWAP_INITIATED:
 		break;
@@ -1035,8 +1161,30 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 		error("Unknown call status %u", status);
 		break;
 	}
+}
 
-	call->status = (int) status;
+static void handle_conference(DBusMessage *msg, gboolean joined)
+{
+	const char *path;
+	struct csd_call *call;
+
+	if (!dbus_message_get_args(msg, NULL,
+					DBUS_TYPE_OBJECT_PATH, &path,
+					DBUS_TYPE_INVALID)) {
+		error("Unexpected parameters in Conference.%s",
+					dbus_message_get_member(msg));
+		return;
+	}
+
+	call = find_call(path);
+	if (!call) {
+		error("Conference signal for unknown call %s", path);
+		return;
+	}
+
+	debug("Call %s %s the conference", path, joined ? "joined" : "left");
+
+	call->conference = joined;
 }
 
 static void get_operator_name_reply(DBusPendingCall *pending_call,
@@ -1051,7 +1199,7 @@ static void get_operator_name_reply(DBusPendingCall *pending_call,
 
 	dbus_error_init(&err);
 	if (dbus_set_error_from_message(&err, reply)) {
-		error("%s get_operator_name failed: %s, %s",
+		error("get_operator_name failed: %s, %s",
 			err.name, err.message);
 		dbus_error_free(&err);
 		goto done;
@@ -1281,19 +1429,16 @@ static void hal_battery_level_reply(DBusPendingCall *call, void *user_data)
 				*value);
 
 	if ((battchg_design > 0 || battchg_last > 0) && battchg_cur >= 0) {
-		int new, cur, max;
+		int new, max;
 
 		if (battchg_last > 0)
 			max = battchg_last;
 		else
 			max = battchg_design;
 
-		cur = telephony_get_indicator(maemo_indicators, "battchg");
 		new = battchg_cur * 5 / max;
 
-		if (new != cur)
-			telephony_update_indicator(maemo_indicators, "battchg",
-							new);
+		telephony_update_indicator(maemo_indicators, "battchg", new);
 	}
 done:
 	dbus_message_unref(reply);
@@ -1361,33 +1506,6 @@ static void handle_hal_property_modified(DBusMessage *msg)
 	}
 }
 
-static DBusHandlerResult signal_filter(DBusConnection *conn,
-						DBusMessage *msg, void *data)
-{
-	const char *path = dbus_message_get_path(msg);
-
-	if (dbus_message_get_type(msg) != DBUS_MESSAGE_TYPE_SIGNAL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-
-	if (dbus_message_is_signal(msg, CSD_CALL_INTERFACE, "Coming"))
-		handle_incoming_call(msg);
-	else if (dbus_message_is_signal(msg, CSD_CALL_INTERFACE, "Created"))
-		handle_outgoing_call(msg);
-	else if (dbus_message_is_signal(msg, CSD_CALL_INSTANCE, "CallStatus"))
-		handle_call_status(msg, path);
-	else if (dbus_message_is_signal(msg, NETWORK_INTERFACE,
-					"registration_status_change"))
-		handle_registration_status_change(msg);
-	else if (dbus_message_is_signal(msg, NETWORK_INTERFACE,
-					"signal_strength_change"))
-		handle_signal_strength_change(msg);
-	else if (dbus_message_is_signal(msg, "org.freedesktop.Hal.Device",
-					"PropertyModified"))
-		handle_hal_property_modified(msg);
-
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-}
-
 static void csd_call_free(struct csd_call *call)
 {
 	if (!call)
@@ -1401,14 +1519,18 @@ static void csd_call_free(struct csd_call *call)
 
 static void parse_call_list(DBusMessageIter *iter)
 {
-	while (dbus_message_iter_get_arg_type(iter)
-						!= DBUS_TYPE_INVALID) {
+	do {
 		DBusMessageIter call_iter;
 		struct csd_call *call;
 		const char *object_path, *number;
 		dbus_uint32_t status;
 		dbus_bool_t originating, terminating, emerg, on_hold, conf;
 
+		if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_STRUCT) {
+			error("Unexpected signature in GetCallInfoAll reply");
+			break;
+		}
+
 		dbus_message_iter_recurse(iter, &call_iter);
 
 		if (!iter_get_basic_args(&call_iter,
@@ -1422,30 +1544,37 @@ static void parse_call_list(DBusMessageIter *iter)
 					DBUS_TYPE_STRING, &number,
 					DBUS_TYPE_INVALID)) {
 			error("Parsing call D-Bus parameters failed");
-			continue;
+			break;
 		}
 
-		call = g_new0(struct csd_call, 1);
+		call = find_call(object_path);
+		if (!call) {
+			call = g_new0(struct csd_call, 1);
+			call->object_path = g_strdup(object_path);
+			call->status = (int) status;
+			calls = g_slist_append(calls, call);
+			debug("telephony-maemo: new csd call instance at %s",
+								object_path);
+		}
 
-		call->object_path = g_strdup(object_path);
-		call->status = (int) status;
-
-		calls = g_slist_append(calls, call);
-
-		debug("telephony-maemo: new csd call instance at %s", object_path);
-
-		if (call->status == CSD_CALL_STATUS_IDLE) {
-			dbus_message_iter_next(iter);
+		if (call->status == CSD_CALL_STATUS_IDLE)
 			continue;
+
+		/* CSD gives incorrect call_hold property sometimes */
+		if ((call->status != CSD_CALL_STATUS_HOLD && on_hold) ||
+				(call->status == CSD_CALL_STATUS_HOLD &&
+								!on_hold)) {
+			error("Conflicting call status and on_hold property!");
+			on_hold = call->status == CSD_CALL_STATUS_HOLD;
 		}
 
 		call->originating = originating;
 		call->on_hold = on_hold;
 		call->conference = conf;
+		g_free(call->number);
 		call->number = g_strdup(number);
 
-		dbus_message_iter_next(iter);
-	}
+	} while (dbus_message_iter_next(iter));
 }
 
 static void signal_strength_reply(DBusPendingCall *call, void *user_data)
@@ -1471,8 +1600,8 @@ static void signal_strength_reply(DBusPendingCall *call, void *user_data)
 					DBUS_TYPE_BYTE, &rssi_in_dbm,
 					DBUS_TYPE_INT32, &net_err,
 					DBUS_TYPE_INVALID)) {
-		error("Unable to parse signal_strength reply:",
-				err.name, err.message);
+		error("Unable to parse signal_strength reply: %s, %s",
+							err.name, err.message);
 		dbus_error_free(&err);
 		return;
 	}
@@ -1504,13 +1633,6 @@ static void registration_status_reply(DBusPendingCall *call, void *user_data)
 	dbus_uint16_t lac, network_type, supported_services;
 	dbus_uint32_t cell_id, operator_code, country_code;
 	dbus_int32_t net_err;
-	uint32_t features = AG_FEATURE_EC_ANDOR_NR |
-				AG_FEATURE_INBAND_RINGTONE |
-				AG_FEATURE_REJECT_A_CALL |
-				AG_FEATURE_ENHANCED_CALL_STATUS |
-				AG_FEATURE_ENHANCED_CALL_CONTROL |
-				AG_FEATURE_EXTENDED_ERROR_RESULT_CODES |
-				AG_FEATURE_THREE_WAY_CALLING;
 
 	reply = dbus_pending_call_steal_reply(call);
 
@@ -1533,8 +1655,8 @@ static void registration_status_reply(DBusPendingCall *call, void *user_data)
 					DBUS_TYPE_BYTE, &supported_services,
 					DBUS_TYPE_INT32, &net_err,
 					DBUS_TYPE_INVALID)) {
-		error("Unable to parse registration_status_change reply:",
-				err.name, err.message);
+		error("Unable to parse registration_status_change reply:"
+					" %s, %s", err.name, err.message);
 		dbus_error_free(&err);
 		return;
 	}
@@ -1548,9 +1670,6 @@ static void registration_status_reply(DBusPendingCall *call, void *user_data)
 					country_code, network_type,
 					supported_services);
 
-	telephony_ready_ind(features, maemo_indicators, response_and_hold,
-				chld_str);
-
 	get_signal_strength();
 
 done:
@@ -1571,6 +1690,8 @@ static void call_info_reply(DBusPendingCall *call, void *user_data)
 	DBusMessage *reply;
 	DBusMessageIter iter, sub;;
 
+	get_calls_active = FALSE;
+
 	reply = dbus_pending_call_steal_reply(call);
 
 	dbus_error_init(&err);
@@ -1620,7 +1741,7 @@ static void hal_find_device_reply(DBusPendingCall *call, void *user_data)
 	dbus_message_iter_init(reply, &iter);
 
 	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_ARRAY) {
-		error("Unexpected signature in GetCallInfoAll return");
+		error("Unexpected signature in FindDeviceByCapability return");
 		goto done;
 	}
 
@@ -1687,6 +1808,8 @@ static void phonebook_read_reply(DBusPendingCall *call, void *user_data)
 
 	if (err != 0) {
 		error("SIM.Phonebook.read failed with error %d", err);
+		if (number_type == &vmbx)
+			vmbx = g_strdup(getenv("VMBX_NUMBER"));
 		goto done;
 	}
 
@@ -1704,49 +1827,21 @@ done:
 	dbus_message_unref(reply);
 }
 
-int telephony_init(void)
+static void csd_init(void)
 {
-	char match_string[128];
-	const char *battery_cap = "battery";
 	dbus_uint32_t location;
 	uint8_t pb_type, location_type;
 	int ret;
 
-	connection = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
-
-	if (!dbus_connection_add_filter(connection, signal_filter,
-						NULL, NULL)) {
-		error("Can't add signal filter");
-		return -EIO;
-	}
-
-	snprintf(match_string, sizeof(match_string),
-			"type=signal,interface=%s", CSD_CALL_INTERFACE);
-	dbus_bus_add_match(connection, match_string, NULL);
-
-	snprintf(match_string, sizeof(match_string),
-			"type=signal,interface=%s", CSD_CALL_INSTANCE);
-	dbus_bus_add_match(connection, match_string, NULL);
-
-	snprintf(match_string, sizeof(match_string),
-			"type=signal,interface=%s", NETWORK_INTERFACE);
-	dbus_bus_add_match(connection, match_string, NULL);
-
 	ret = send_method_call(CSD_CALL_BUS_NAME, CSD_CALL_PATH,
 				CSD_CALL_INTERFACE, "GetCallInfoAll",
 				call_info_reply, NULL, DBUS_TYPE_INVALID);
-	if (ret < 0)
-		return ret;
+	if (ret < 0) {
+		error("Unable to sent GetCallInfoAll method call");
+		return;
+	}
 
-	ret = send_method_call("org.freedesktop.Hal",
-				"/org/freedesktop/Hal/Manager",
-				"org.freedesktop.Hal.Manager",
-				"FindDeviceByCapability",
-				hal_find_device_reply, NULL,
-				DBUS_TYPE_STRING, &battery_cap,
-				DBUS_TYPE_INVALID);
-	if (ret < 0)
-		return ret;
+	get_calls_active = TRUE;
 
 	pb_type = SIM_PHONEBOOK_TYPE_MSISDN;
 	location = PHONEBOOK_INDEX_FIRST_ENTRY;
@@ -1759,20 +1854,251 @@ int telephony_init(void)
 				DBUS_TYPE_INT32, &location,
 				DBUS_TYPE_BYTE, &location_type,
 				DBUS_TYPE_INVALID);
-	if (ret < 0)
-		return ret;
+	if (ret < 0) {
+		error("Unable to send " SIM_PHONEBOOK_INTERFACE ".read()");
+		return;
+	}
 
-	pb_type = SIM_PHONEBOOK_TYPE_VMBX;
+	pb_type = SIM_PHONEBOOK_TYPE_MBDN;
 	location = PHONEBOOK_INDEX_FIRST_ENTRY;
 	location_type = SIM_PHONEBOOK_LOCATION_NEXT;
 
-	return send_method_call(SIM_PHONEBOOK_BUS_NAME, SIM_PHONEBOOK_PATH,
+	ret = send_method_call(SIM_PHONEBOOK_BUS_NAME, SIM_PHONEBOOK_PATH,
 				SIM_PHONEBOOK_INTERFACE, "read",
 				phonebook_read_reply, &vmbx,
 				DBUS_TYPE_BYTE, &pb_type,
 				DBUS_TYPE_INT32, &location,
 				DBUS_TYPE_BYTE, &location_type,
 				DBUS_TYPE_INVALID);
+	if (ret < 0) {
+		error("Unable to send " SIM_PHONEBOOK_INTERFACE ".read()");
+		return;
+	}
+}
+
+static inline DBusMessage *invalid_args(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg,"org.bluez.Error.InvalidArguments",
+					"Invalid arguments in method call");
+}
+
+static uint32_t get_callflag(const char *callerid_setting)
+{
+	if (callerid_setting != NULL) {
+		if (g_str_equal(callerid_setting, "allowed"))
+			return CALL_FLAG_PRESENTATION_ALLOWED;
+		else if (g_str_equal(callerid_setting, "restricted"))
+			return CALL_FLAG_PRESENTATION_RESTRICTED;
+		else
+			return CALL_FLAG_NONE;
+	} else
+		return CALL_FLAG_NONE;
+}
+
+static void generate_flag_file(const char *filename)
+{
+	int fd;
+
+	if (g_file_test(ALLOWED_FLAG_FILE, G_FILE_TEST_EXISTS) ||
+			g_file_test(RESTRICTED_FLAG_FILE, G_FILE_TEST_EXISTS) ||
+			g_file_test(NONE_FLAG_FILE, G_FILE_TEST_EXISTS))
+		return;
+
+	fd = open(filename, O_WRONLY | O_CREAT, 0);
+	if (fd >= 0)
+		close(fd);
+}
+
+static void save_callerid_to_file(const char *callerid_setting)
+{
+	char callerid_file[FILENAME_MAX];
+
+	snprintf(callerid_file, sizeof(callerid_file), "%s%s",
+					CALLERID_BASE, callerid_setting);
+
+	if (g_file_test(ALLOWED_FLAG_FILE, G_FILE_TEST_EXISTS))
+		rename(ALLOWED_FLAG_FILE, callerid_file);
+	else if (g_file_test(RESTRICTED_FLAG_FILE, G_FILE_TEST_EXISTS))
+		rename(RESTRICTED_FLAG_FILE, callerid_file);
+	else if (g_file_test(NONE_FLAG_FILE, G_FILE_TEST_EXISTS))
+		rename(NONE_FLAG_FILE, callerid_file);
+	else
+		generate_flag_file(callerid_file);
+}
+
+static uint32_t callerid_from_file(void)
+{
+	if (g_file_test(ALLOWED_FLAG_FILE, G_FILE_TEST_EXISTS))
+		return CALL_FLAG_PRESENTATION_ALLOWED;
+	else if (g_file_test(RESTRICTED_FLAG_FILE, G_FILE_TEST_EXISTS))
+		return CALL_FLAG_PRESENTATION_RESTRICTED;
+	else if (g_file_test(NONE_FLAG_FILE, G_FILE_TEST_EXISTS))
+		return CALL_FLAG_NONE;
+	else
+		return CALL_FLAG_NONE;
+}
+
+static DBusMessage *set_callerid(DBusConnection *conn, DBusMessage *msg,
+					void *data)
+{
+	const char *callerid_setting;
+
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING,
+						&callerid_setting,
+						DBUS_TYPE_INVALID) == FALSE)
+		return invalid_args(msg);
+
+	if (g_str_equal(callerid_setting, "allowed") ||
+			g_str_equal(callerid_setting, "restricted") ||
+			g_str_equal(callerid_setting, "none")) {
+		save_callerid_to_file(callerid_setting);
+		callerid = get_callflag(callerid_setting);
+		debug("telephony-maemo setting callerid flag: %s",
+							callerid_setting);
+		return dbus_message_new_method_return(msg);
+	}
+
+	error("telephony-maemo: invalid argument %s for method call"
+					" SetCallerId", callerid_setting);
+		return invalid_args(msg);
+}
+
+static GDBusMethodTable telephony_maemo_methods[] = {
+	{"SetCallerId",		"s",	"",	set_callerid,
+						G_DBUS_METHOD_FLAG_ASYNC},
+	{ }
+};
+
+static void handle_modem_state(DBusMessage *msg)
+{
+	const char *state;
+
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &state,
+							DBUS_TYPE_INVALID)) {
+		error("Unexpected modem state parameters");
+		return;
+	}
+
+	debug("SSC modem state: %s", state);
+
+	if (calls != NULL || get_calls_active)
+		return;
+
+	if (g_str_equal(state, "cmt_ready") || g_str_equal(state, "online"))
+		csd_init();
+}
+
+static void modem_state_reply(DBusPendingCall *call, void *user_data)
+{
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+	DBusError err;
+
+	dbus_error_init(&err);
+	if (dbus_set_error_from_message(&err, reply)) {
+		error("get_modem_status: %s, %s", err.name, err.message);
+		dbus_error_free(&err);
+	} else
+		handle_modem_state(reply);
+
+	dbus_message_unref(reply);
+}
+
+static DBusHandlerResult signal_filter(DBusConnection *conn,
+						DBusMessage *msg, void *data)
+{
+	const char *path = dbus_message_get_path(msg);
+
+	if (dbus_message_get_type(msg) != DBUS_MESSAGE_TYPE_SIGNAL)
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+	if (dbus_message_is_signal(msg, CSD_CALL_INTERFACE, "Coming"))
+		handle_incoming_call(msg);
+	else if (dbus_message_is_signal(msg, CSD_CALL_INTERFACE, "Created"))
+		handle_outgoing_call(msg);
+	else if (dbus_message_is_signal(msg, CSD_CALL_INTERFACE,
+							"CreateRequested"))
+		handle_create_requested(msg);
+	else if (dbus_message_is_signal(msg, CSD_CALL_INSTANCE, "CallStatus"))
+		handle_call_status(msg, path);
+	else if (dbus_message_is_signal(msg, CSD_CALL_CONFERENCE, "Joined"))
+		handle_conference(msg, TRUE);
+	else if (dbus_message_is_signal(msg, CSD_CALL_CONFERENCE, "Left"))
+		handle_conference(msg, FALSE);
+	else if (dbus_message_is_signal(msg, NETWORK_INTERFACE,
+					"registration_status_change"))
+		handle_registration_status_change(msg);
+	else if (dbus_message_is_signal(msg, NETWORK_INTERFACE,
+					"signal_strength_change"))
+		handle_signal_strength_change(msg);
+	else if (dbus_message_is_signal(msg, "org.freedesktop.Hal.Device",
+					"PropertyModified"))
+		handle_hal_property_modified(msg);
+	else if (dbus_message_is_signal(msg, SSC_DBUS_IFACE,
+						"modem_state_changed_ind"))
+		handle_modem_state(msg);
+
+	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
+int telephony_init(void)
+{
+	const char *battery_cap = "battery";
+	uint32_t features = AG_FEATURE_EC_ANDOR_NR |
+				AG_FEATURE_INBAND_RINGTONE |
+				AG_FEATURE_REJECT_A_CALL |
+				AG_FEATURE_ENHANCED_CALL_STATUS |
+				AG_FEATURE_ENHANCED_CALL_CONTROL |
+				AG_FEATURE_EXTENDED_ERROR_RESULT_CODES |
+				AG_FEATURE_THREE_WAY_CALLING;
+
+	connection = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
+
+	if (!dbus_connection_add_filter(connection, signal_filter,
+						NULL, NULL))
+		error("Can't add signal filter");
+
+	dbus_bus_add_match(connection,
+			"type=signal,interface=" CSD_CALL_INTERFACE, NULL);
+	dbus_bus_add_match(connection,
+			"type=signal,interface=" CSD_CALL_INSTANCE, NULL);
+	dbus_bus_add_match(connection,
+			"type=signal,interface=" CSD_CALL_CONFERENCE, NULL);
+	dbus_bus_add_match(connection,
+			"type=signal,interface=" NETWORK_INTERFACE, NULL);
+	dbus_bus_add_match(connection,
+				"type=signal,interface=" SSC_DBUS_IFACE
+				",member=modem_state_changed_ind", NULL);
+
+	if (send_method_call(SSC_DBUS_NAME, SSC_DBUS_PATH, SSC_DBUS_IFACE,
+					"get_modem_state", modem_state_reply,
+					NULL, DBUS_TYPE_INVALID) < 0)
+		error("Unable to send " SSC_DBUS_IFACE ".get_modem_state()");
+
+	generate_flag_file(NONE_FLAG_FILE);
+	callerid = callerid_from_file();
+
+	if (!g_dbus_register_interface(connection, TELEPHONY_MAEMO_PATH,
+			TELEPHONY_MAEMO_INTERFACE, telephony_maemo_methods,
+			NULL, NULL, NULL, NULL)) {
+		error("telephony-maemo interface %s init failed on path %s",
+			TELEPHONY_MAEMO_INTERFACE, TELEPHONY_MAEMO_PATH);
+	}
+
+	debug("telephony-maemo registering %s interface on path %s",
+			TELEPHONY_MAEMO_INTERFACE, TELEPHONY_MAEMO_PATH);
+
+	telephony_ready_ind(features, maemo_indicators, response_and_hold,
+								chld_str);
+	if (send_method_call("org.freedesktop.Hal",
+				"/org/freedesktop/Hal/Manager",
+				"org.freedesktop.Hal.Manager",
+				"FindDeviceByCapability",
+				hal_find_device_reply, NULL,
+				DBUS_TYPE_STRING, &battery_cap,
+				DBUS_TYPE_INVALID) < 0)
+		error("Unable to send HAL method call");
+
+	return 0;
 }
 
 void telephony_exit(void)
