@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2001-2002  Nokia Corporation
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2010  Marcel Holtmann <marcel@holtmann.org>
  *  Copyright (C) 2002-2003  Stephen Crane <steve.crane@rococosoft.com>
  *
  *
@@ -166,7 +166,8 @@ uint8_t get_service_classes(const bdaddr_t *bdaddr)
 	return service_classes;
 }
 
-void create_ext_inquiry_response(const char *name, uint8_t *data)
+void create_ext_inquiry_response(const char *name,
+					int8_t tx_power, uint8_t *data)
 {
 	sdp_list_t *list = sdp_get_record_list();
 	uint8_t *ptr = data;
@@ -189,6 +190,12 @@ void create_ext_inquiry_response(const char *name, uint8_t *data)
 		ptr += len + 2;
 	}
 
+	if (tx_power != 0) {
+		*ptr++ = 2;
+		*ptr++ = 0x0a;
+		*ptr++ = (uint8_t) tx_power;
+	}
+
 	if (did_vendor != 0x0000) {
 		uint16_t source = 0x0002;
 		*ptr++ = 9;
@@ -214,6 +221,9 @@ void create_ext_inquiry_response(const char *name, uint8_t *data)
 		if (rec->svclass.value.uuid16 < 0x1100)
 			continue;
 
+		if (rec->svclass.value.uuid16 == PNP_INFO_SVCLASS_ID)
+			continue;
+
 		if (index > 23) {
 			ptr[1] = 0x02;
 			break;
