@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2000-2001  Qualcomm Incorporated
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -980,7 +980,7 @@ static void cmd_inq_tpl(int ctl, int hdev, char *opt)
 	} else {
 		int8_t level;
 
-		if (hci_read_inquiry_transmit_power_level(dd, &level, 1000) < 0) {
+		if (hci_read_inq_response_tx_power_level(dd, &level, 1000) < 0) {
 			fprintf(stderr, "Can't read inquiry transmit power level on hci%d: %s (%d)\n",
 						hdev, strerror(errno), errno);
 			exit(1);
@@ -1127,7 +1127,11 @@ static void cmd_inq_data(int ctl, int hdev, char *opt)
 				}
 				break;
 			case 0x0a:
-				printf("\tTX power level: %d\n", *((uint8_t *) ptr));
+				printf("\tTX power level: %d\n", *((int8_t *) ptr));
+				break;
+			case 0x10:
+				printf("\tDevice ID with %d bytes data\n",
+								len - 1);
 				break;
 			default:
 				printf("\tUnknown type 0x%02x with %d bytes data\n",
